from __future__ import annotations

import sys
from typing import TYPE_CHECKING

from django.core.management import ManagementUtility as BaseManagementUtility
from django.core.management.color import color_style

from .conf import settings
from .core import import_command_class, load_command_class

if TYPE_CHECKING:
    from django.core.management.base import BaseCommand

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override


class ManagementUtility(BaseManagementUtility):
    @override
    def main_help_text(self, commands_only: bool = False) -> str:
        usage = super().main_help_text(commands_only=commands_only)

        style = color_style()

        commands_usage = (
            [
                style.NOTICE("[django-management-commands: paths]"),
                *[f"    {path}" for path in paths],
                "",
            ]
            if (paths := settings.PATHS)
            else []
        )
        aliases_usage = (
            [
                style.NOTICE("[django-management-commands: aliases]"),
                *[f"    {alias}" for alias in aliases],
                "",
            ]
            if (aliases := settings.ALIASES)
            else []
        )

        usage_list = usage.split("\n")
        usage_list.append("")
        usage_list.extend(commands_usage)
        usage_list.extend(aliases_usage)

        return "\n".join(usage_list)

    @override
    def fetch_command(self, subcommand: str) -> BaseCommand:
        if dotted_path := settings.PATHS.get(subcommand):
            command_class = import_command_class(dotted_path)
        else:
            try:
                app_label, name = subcommand.rsplit(".", 1)
            except ValueError:
                app_label, name = None, subcommand

            command_class = load_command_class(name, app_label)

        return command_class()

    @override
    def execute(self) -> None:
        try:
            name = self.argv[1]
        except IndexError:
            super().execute()
        else:
            if name in settings.PATHS:
                utility = self.__class__([self.prog_name, name, *self.argv[2:]])
                super(ManagementUtility, utility).execute()
            elif alias_exprs := settings.ALIASES.get(name):
                for alias_expr in alias_exprs:
                    argv = alias_expr.split()

                    utility = ManagementUtility([self.prog_name, *argv])
                    utility.execute()
            else:
                super().execute()


def execute_from_command_line(argv: list[str] | None = None) -> None:
    utility = ManagementUtility(argv)
    utility.execute()



from __future__ import annotations

from django.core.management.base import BaseCommand


class ManagementCommandsException(Exception):
    msg: str

    def __init__(self, msg: str | None = None, **kwargs: str) -> None:
        super().__init__(
            msg
            if msg
            else (
                msg_template.format(**kwargs)
                if (msg_template := getattr(type(self), "msg", None)) and kwargs
                else ""
            ),
        )


class CommandClassLookupError(ManagementCommandsException):
    msg = "command {command_name!r} is not registered from {app_info}"

    def __init__(
        self,
        command_name: str | None = None,
        app_name: str | None = None,
    ) -> None:
        super().__init__(
            **(
                {
                    "command_name": command_name,
                    "app_info": (
                        f"the {app_name!r} app"
                        if app_name
                        else "any of the installed apps"
                    ),
                }
                if command_name
                else {}
            ),
        )


class CommandAppLookupError(ManagementCommandsException):
    msg = "app {app_name!r} is not installed"

    def __init__(self, app_name: str | None = None) -> None:
        super().__init__(**({"app_name": app_name} if app_name else {}))


class CommandImportError(ManagementCommandsException, ImportError):
    msg = "class {class_name!r} could not be imported from the {module_path!r} module"

    def __init__(self, command_class_path: str | None = None) -> None:
        super().__init__(
            **({"module_path": parts[0], "class_name": parts[1]})
            if (command_class_path and (parts := command_class_path.rsplit(".", 1)))
            else {},
        )


class CommandTypeError(ManagementCommandsException):
    msg = "class {class_path!r} is not a subclass of {base_command_class_path!r}"

    def __init__(self, class_: type | None = None) -> None:
        super().__init__(
            **(
                {
                    "class_path": self._class_path(class_),
                    "base_command_class_path": self._class_path(BaseCommand),
                }
                if class_
                else {}
            ),
        )

    @staticmethod
    def _class_path(class_: type) -> str:
        return f"{class_.__module__}.{class_.__qualname__}"



from __future__ import annotations

from typing import TYPE_CHECKING, cast

import pytest

from django.core.management import get_commands
from django.core.management.base import BaseCommand

from management_commands.management import execute_from_command_line

if TYPE_CHECKING:
    from pytest_mock import MockerFixture


@pytest.fixture(params=get_commands())
def django_core_command_name(request: pytest.FixtureRequest) -> str:
    return cast(str, request.param)


def test_execute_from_command_line_help_displays_paths_and_aliases(
    mocker: MockerFixture,
    capsys: pytest.CaptureFixture[str],
) -> None:
    # Mock.
    mocker.patch.multiple(
        "management_commands.management.settings",
        PATHS={
            "command_a": "module.CommandA",
            "command_b": "module.CommandB",
        },
        ALIASES={
            "alias_a": [
                "command_a",
            ],
            "alias_b": [
                "command_b",
            ],
        },
    )

    # Act.
    execute_from_command_line(["manage.py", "--help"])
    captured = capsys.readouterr()

    # Assert.
    assert (
        "[django-management-commands: paths]\n"
        "    command_a\n"
        "    command_b\n"
        "\n"
        "[django-management-commands: aliases]\n"
        "    alias_a\n"
        "    alias_b\n"
        "\n"
    ) in captured.out


def test_execute_from_command_line_falls_back_to_django_management_utility_if_command_name_is_not_passed(
    mocker: MockerFixture,
) -> None:
    # Mock.
    super_mock = mocker.patch("management_commands.management.super")

    # Act.
    execute_from_command_line(["manage.py"])

    # Assert.
    super_mock.return_value.execute.assert_called_once()


def test_execute_from_command_line_falls_back_to_django_management_utility_to_run_command_not_defined_in_settings(
    mocker: MockerFixture,
) -> None:
    # Configure.
    mocker.patch.multiple(
        "management_commands.management.settings",
        PATHS={},
        ALIASES={},
    )

    # Mock.
    super_mock = mocker.patch("management_commands.management.super")

    # Act.
    execute_from_command_line(["manage.py", "command"])

    # Assert.
    super_mock.return_value.execute.assert_called_once()


def test_execute_from_command_line_runs_command_from_path(
    mocker: MockerFixture,
) -> None:
    # Configure.
    mocker.patch(
        "management_commands.management.settings.PATHS",
        {
            "command": "module.Command",
        },
    )

    # Arrange.
    class Command(BaseCommand):
        pass

    # Mock.
    def import_string_side_effect(dotted_path: str) -> type[BaseCommand]:
        if dotted_path == "module.Command":
            return Command
        raise ImportError

    mocker.patch(
        "management_commands.core.import_string",
        side_effect=import_string_side_effect,
    )

    command_run_from_argv_mock = mocker.patch.object(Command, "run_from_argv")

    # Act.
    execute_from_command_line(["manage.py", "command"])

    # Assert.
    command_run_from_argv_mock.assert_called_once()


def test_execute_from_command_line_uses_django_management_utility_to_run_command_from_path(
    mocker: MockerFixture,
) -> None:
    # Configure.
    mocker.patch(
        "management_commands.management.settings.PATHS",
        {
            "command": "module.Command",
        },
    )

    # Mock.
    super_mock = mocker.patch("management_commands.management.super")

    # Act.
    execute_from_command_line(["manage.py", "command"])

    # Assert.
    super_mock.return_value.execute.assert_called_once()


def test_execute_from_command_lines_runs_all_commands_assigned_to_alias(
    mocker: MockerFixture,
) -> None:
    # Configure.
    mocker.patch(
        "management_commands.management.settings.ALIASES",
        {
            "alias": [
                "command_a arg_a --option value_a",
                "command_b arg_b --option value_b",
            ],
        },
    )

    # Arrange.
    class CommandA(BaseCommand):
        pass

    class CommandB(BaseCommand):
        pass

    app_config_mock = mocker.Mock()
    app_config_mock.name = "app"

    # Mock.
    def import_string_side_effect(dotted_path: str) -> type[BaseCommand]:
        if dotted_path == "app.management.commands.command_a.Command":
            return CommandA
        if dotted_path == "app.management.commands.command_b.Command":
            return CommandB
        raise ImportError

    mocker.patch(
        "management_commands.core.apps.app_configs",
        {
            "app": app_config_mock,
        },
    )
    mocker.patch(
        "management_commands.core.import_string",
        side_effect=import_string_side_effect,
    )

    command_a_run_from_argv_mock = mocker.patch.object(
        CommandA,
        "run_from_argv",
        return_value=None,
    )
    command_b_run_from_argv_mock = mocker.patch.object(
        CommandB,
        "run_from_argv",
        return_value=None,
    )

    # Act.
    execute_from_command_line(["manage.py", "alias"])

    # Assert.
    command_a_run_from_argv_mock.assert_called_once_with(
        ["manage.py", "command_a", "arg_a", "--option", "value_a"],
    )
    command_b_run_from_argv_mock.assert_called_once_with(
        ["manage.py", "command_b", "arg_b", "--option", "value_b"],
    )


def test_execute_from_command_line_prefers_path_command_over_django_core_command(
    mocker: MockerFixture,
    django_core_command_name: str,
) -> None:
    # Configure.
    mocker.patch(
        "management_commands.management.settings.PATHS",
        {
            django_core_command_name: "module.Command",
        },
    )

    # Arrange.
    class Command(BaseCommand):
        pass

    # Mock.
    def import_string_side_effect(dotted_path: str) -> type[BaseCommand]:
        if dotted_path == "module.Command":
            return Command
        raise ImportError

    mocker.patch(
        "management_commands.core.import_string",
        side_effect=import_string_side_effect,
    )

    command_run_from_argv_mock = mocker.patch.object(Command, "run_from_argv")

    # Act.
    execute_from_command_line(["manage.py", django_core_command_name])

    # Assert.
    command_run_from_argv_mock.assert_called_once()


def test_execute_from_command_line_prefers_path_command_over_alias(
    mocker: MockerFixture,
) -> None:
    # Configure.
    mocker.patch.multiple(
        "management_commands.management.settings",
        PATHS={
            "command": "module.CommandA",
        },
        ALIASES={
            "command": [
                "command_b",
            ],
        },
    )

    # Arrange.
    class CommandA(BaseCommand):
        pass

    class CommandB(BaseCommand):
        pass

    app_config_mock = mocker.Mock()
    app_config_mock.name = "app"

    # Mock.
    def import_string_side_effect(dotted_path: str) -> type[BaseCommand]:
        if dotted_path == "module.CommandA":
            return CommandA
        if dotted_path == "app.management.commands.command_b.Command":
            return CommandB
        raise ImportError

    mocker.patch(
        "management_commands.core.apps.app_configs",
        {
            "app": app_config_mock,
        },
    )
    mocker.patch(
        "management_commands.core.import_string",
        side_effect=import_string_side_effect,
    )

    command_a_run_from_argv_mock = mocker.patch.object(CommandA, "run_from_argv")

    # Act.
    execute_from_command_line(["manage.py", "command"])

    # Assert.
    command_a_run_from_argv_mock.assert_called_once()


def test_execute_from_command_line_prefers_alias_over_django_core_command(
    mocker: MockerFixture,
    django_core_command_name: str,
) -> None:
    # Configure.
    mocker.patch.multiple(
        "management_commands.management.settings",
        ALIASES={
            django_core_command_name: [
                "command",
            ],
        },
    )

    # Arrange.
    class Command(BaseCommand):
        pass

    class DjangoCoreCommand(BaseCommand):
        pass

    app_config_mock = mocker.Mock()
    app_config_mock.name = "app"

    # Mock.
    def import_string_side_effect(dotted_path: str) -> type[BaseCommand]:
        if dotted_path == "app.management.commands.command.Command":
            return Command
        if (
            dotted_path
            == f"django.core.management.commands.{django_core_command_name}.Command"
        ):
            return DjangoCoreCommand
        raise ImportError

    mocker.patch(
        "management_commands.core.apps.app_configs",
        {
            "app": app_config_mock,
        },
    )
    mocker.patch(
        "management_commands.core.import_string",
        side_effect=import_string_side_effect,
    )

    command_run_from_argv_mock = mocker.patch.object(Command, "run_from_argv")

    # Act.
    execute_from_command_line(["manage.py", django_core_command_name])

    # Assert.
    command_run_from_argv_mock.assert_called_once()


def test_execute_from_command_line_resolves_django_core_commands_and_aliases_if_app_label_is_provided(
    mocker: MockerFixture,
    django_core_command_name: str,
) -> None:
    # Configure.
    mocker.patch.multiple(
        "management_commands.management.settings",
        ALIASES={
            "alias": [
                f"app.{django_core_command_name}",
            ],
        },
    )

    # Arrange.
    class Command(BaseCommand):
        pass

    class DjangoCoreCommand(BaseCommand):
        pass

    app_config_mock = mocker.Mock()
    app_config_mock.name = "app"

    # Mock.
    def import_string_side_effect(dotted_path: str) -> type[BaseCommand]:
        if dotted_path == f"app.management.commands.{django_core_command_name}.Command":
            return Command
        if (
            dotted_path
            == f"django.core.management.commands.{django_core_command_name}.Command"
        ):
            return DjangoCoreCommand
        raise ImportError

    mocker.patch(
        "management_commands.core.apps.app_configs",
        {
            "app": app_config_mock,
        },
    )
    mocker.patch(
        "management_commands.core.import_string",
        side_effect=import_string_side_effect,
    )

    command_run_from_argv_mock = mocker.patch.object(Command, "run_from_argv")

    # Act.
    execute_from_command_line(["manage.py", "alias"])

    # Assert.
    command_run_from_argv_mock.assert_called_once()


def test_execute_from_command_line_runs_command_passed_with_explicit_app_label(
    mocker: MockerFixture,
) -> None:
    # Arrange.
    class Command(BaseCommand):
        pass

    app_config_mock = mocker.Mock()
    app_config_mock.name = "app"

    # Mock.
    def import_string_side_effect(dotted_path: str) -> type[BaseCommand]:
        if dotted_path == "app.management.commands.command.Command":
            return Command
        raise ImportError

    mocker.patch(
        "management_commands.core.apps.app_configs",
        {
            "app": app_config_mock,
        },
    )
    mocker.patch(
        "management_commands.core.import_string",
        side_effect=import_string_side_effect,
    )

    command_run_from_argv_mock = mocker.patch.object(Command, "run_from_argv")

    # Act.
    execute_from_command_line(["manage.py", "app.command"])

    # Assert.
    command_run_from_argv_mock.assert_called_once()


def test_execute_from_command_line_runs_command_defined_in_path_when_referenced_by_alias(
    mocker: MockerFixture,
) -> None:
    # Configure.
    mocker.patch.multiple(
        "management_commands.management.settings",
        PATHS={
            "command": "module.Command",
        },
        ALIASES={
            "alias": [
                "command",
            ],
        },
    )

    # Arrange.
    class Command(BaseCommand):
        pass

    # Mock.
    def import_string_side_effect(dotted_path: str) -> type[BaseCommand]:
        if dotted_path == "module.Command":
            return Command
        raise ImportError

    mocker.patch(
        "management_commands.core.import_string",
        side_effect=import_string_side_effect,
    )

    command_run_from_argv_mock = mocker.patch.object(Command, "run_from_argv")

    # Act.
    execute_from_command_line(["manage.py", "alias"])

    # Assert.
    command_run_from_argv_mock.assert_called_once()


def test_execute_from_command_line_runs_command_defined_in_alias_via_other_alias(
    mocker: MockerFixture,
) -> None:
    # Configure.
    mocker.patch.multiple(
        "management_commands.management.settings",
        ALIASES={
            "alias_a": [
                "command",
            ],
            "alias_b": [
                "alias_a",
            ],
        },
    )

    # Arrange.
    class Command(BaseCommand):
        pass

    app_config_mock = mocker.Mock()
    app_config_mock.name = "app"

    # Mock.
    def import_string_side_effect(dotted_path: str) -> type[BaseCommand]:
        if dotted_path == "app.management.commands.command.Command":
            return Command
        raise ImportError

    mocker.patch(
        "management_commands.core.apps.app_configs",
        {
            "app": app_config_mock,
        },
    )
    mocker.patch(
        "management_commands.core.import_string",
        side_effect=import_string_side_effect,
    )

    command_run_from_argv_mock = mocker.patch.object(Command, "run_from_argv")

    # Act.
    execute_from_command_line(["manage.py", "alias_b"])

    # Assert.
    command_run_from_argv_mock.assert_called_once()


def test_execute_from_command_line_resolves_duplicate_command_names_based_on_order_of_installed_apps(
    mocker: MockerFixture,
) -> None:
    # Configure.
    mocker.patch.multiple(
        "management_commands.management.settings",
        ALIASES={
            "alias": [
                "command",
            ],
        },
    )

    # Arrange.
    class CommandA(BaseCommand):
        pass

    class CommandB(BaseCommand):
        pass

    app_config_a_mock = mocker.Mock()
    app_config_a_mock.name = "app_a"
    app_config_b_mock = mocker.Mock()
    app_config_b_mock.name = "app_b"

    # Mock.
    def import_string_side_effect(dotted_path: str) -> type[BaseCommand]:
        if dotted_path == "app_a.management.commands.command.Command":
            return CommandA
        if dotted_path == "app_b.management.commands.command.Command":
            return CommandB
        raise ImportError

    mocker.patch(
        "management_commands.core.apps.app_configs",
        {
            "app_a": app_config_a_mock,
            "app_b": app_config_b_mock,
        },
    )
    mocker.patch(
        "management_commands.core.import_string",
        side_effect=import_string_side_effect,
    )

    command_b_run_from_argv_mock = mocker.patch.object(CommandB, "run_from_argv")

    # Act.
    execute_from_command_line(["manage.py", "alias"])

    # Assert.
    command_b_run_from_argv_mock.assert_called_once()



from __future__ import annotations

import re
from keyword import iskeyword
from typing import ClassVar

import appconf


def _is_identifier(s: str) -> bool:
    return s.isidentifier() and not iskeyword(s)


def _is_dotted_path(s: str, /, *, min_parts: int = 0) -> bool:
    if not (dotted_path_match := re.match(r"^[^\d\W]\w*(\.[^\d\W]\w*)*$", s)):
        return False

    parts = dotted_path_match.group().split(".")

    return len(parts) >= min_parts and all(map(_is_identifier, parts))


class ManagementCommandsConf(appconf.AppConf):  # type: ignore[misc]
    PATHS: ClassVar[dict[str, str]] = {}

    MODULES: ClassVar[list[str]] = []

    SUBMODULES: ClassVar[list[str]] = []

    ALIASES: ClassVar[dict[str, list[str]]] = {}

    class ImproperlyConfigured(Exception):
        def __init__(self, msg: str, code: str | None = None) -> None:
            super().__init__(msg)

            self.code = code

    def improperly_configured(
        self,
        msg: str,
        code: str | None = None,
    ) -> ImproperlyConfigured:
        msg = re.sub(
            rf"({'|'.join(dir(self))})",
            lambda m: f"{self._meta.prefixed_name(m.group(0))}",
            msg,
        )

        return self.__class__.ImproperlyConfigured(msg, code)

    def configure_paths(self, setting_value: dict[str, str]) -> dict[str, str]:
        for key, value in setting_value.items():
            if not _is_identifier(key):
                msg = (
                    f"invalid key {key!r} in PATHS; "
                    f"keys must be valid Python identifiers"
                )

                raise self.improperly_configured(msg, "paths.key")

            if not _is_dotted_path(value, min_parts=2):
                msg = (
                    f"invalid value for PATHS[{key!r}]; "
                    f"values must be valid absolute dotted paths with at least 2 parts"
                )

                raise self.improperly_configured(msg, "paths.value")

        return setting_value

    def _configure_path_list(
        self,
        setting_name: str,
        setting_value: list[str],
    ) -> list[str]:
        for index, item in enumerate(setting_value):
            if not _is_dotted_path(item):
                msg = (
                    f"invalid value for {setting_name.upper()}[{index}]; "
                    f"items must be valid absolute dotted paths"
                )

                raise self.improperly_configured(msg, f"{setting_name}.item")

        return setting_value

    def configure_modules(self, setting_value: list[str]) -> list[str]:
        return self._configure_path_list("modules", setting_value)

    def configure_submodules(self, setting_value: list[str]) -> list[str]:
        configured_value = self._configure_path_list("submodules", setting_value)

        if "management.commands" not in configured_value:
            configured_value.insert(0, "management.commands")

        return configured_value

    def configure_aliases(
        self,
        setting_value: dict[str, list[str]],
    ) -> dict[str, list[str]]:
        for key, value in setting_value.items():
            if not _is_identifier(key):
                msg = (
                    f"invalid key {key!r} in ALIASES; "
                    f"keys must be valid Python identifiers"
                )

                raise self.improperly_configured(msg, "aliases.key")

            for index, item in enumerate(value):
                argv = item.split()

                try:
                    command = argv[0]
                except IndexError as exc:
                    msg = (
                        f"empty item found in ALIASES[{key!r}][{index}]; "
                        f"items must not be empty"
                    )

                    raise self.improperly_configured(msg, "aliases.empty") from exc

                if command == key:
                    msg = (
                        f"invalid value for ALIASES[{key!r}][{index}]; "
                        f"items must not refer to the aliases they are defined by"
                    )

                    raise self.improperly_configured(msg, "aliases.self_reference")

        return setting_value


settings = ManagementCommandsConf()



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from django.core.management.base import BaseCommand

from management_commands.core import (
    get_command_paths,
    import_command_class,
    load_command_class,
)
from management_commands.exceptions import (
    CommandAppLookupError,
    CommandClassLookupError,
    CommandImportError,
    CommandTypeError,
)

if TYPE_CHECKING:
    from pytest_mock import MockerFixture


def test_import_command_class_imports_command_if_dotted_path_points_to_base_command_subclass(
    mocker: MockerFixture,
) -> None:
    # Arrange.
    class Command(BaseCommand):
        pass

    # Mock.
    mocker.patch(
        "management_commands.core.import_string",
        return_value=Command,
    )

    # Act.
    command_class = import_command_class("module.Command")

    # Assert.
    assert command_class is Command


def test_import_command_class_raises_command_import_error_if_importing_from_dotted_path_fails(
    mocker: MockerFixture,
) -> None:
    # Mock.
    mocker.patch(
        "management_commands.core.import_string",
        side_effect=ImportError,
    )

    # Act & assert.
    with pytest.raises(CommandImportError):
        import_command_class("module_does_not_exist.Command")


def test_import_command_class_raises_command_type_error_if_imported_class_is_not_a_subclass_of_base_command(
    mocker: MockerFixture,
) -> None:
    # Arrange.
    class NotABaseCommand:
        pass

    # Mock.
    mocker.patch(
        "management_commands.core.import_string",
        return_value=NotABaseCommand,
    )

    # Act & assert.
    with pytest.raises(CommandTypeError):
        import_command_class("module.Command")


def test_get_command_paths_returns_list_of_all_dotted_paths_to_command_classes_in_modules_and_submodules_if_app_label_unspecified(
    mocker: MockerFixture,
) -> None:
    # Configure.
    mocker.patch.multiple(
        "management_commands.conf.settings",
        MODULES=[
            "module_a",
            "module_b",
        ],
        SUBMODULES=[
            "submodule_a",
            "submodule_b",
        ],
    )

    # Arrange.
    app_config_a_mock = mocker.Mock()
    app_config_a_mock.name = "app_a"
    app_config_b_mock = mocker.Mock()
    app_config_b_mock.name = "app_b"

    # Mock.
    mocker.patch(
        "management_commands.core.apps.app_configs",
        {
            "app_a": app_config_a_mock,
            "app_b": app_config_b_mock,
        },
    )

    # Act.
    command_paths = get_command_paths("command")

    # Assert.
    assert command_paths == [
        "module_a.command.Command",
        "module_b.command.Command",
        "app_b.submodule_a.command.Command",
        "app_b.submodule_b.command.Command",
        "app_a.submodule_a.command.Command",
        "app_a.submodule_b.command.Command",
    ]


def test_get_command_paths_returns_list_of_dotted_paths_to_app_submodules_if_app_label_specified(
    mocker: MockerFixture,
) -> None:
    # Configure.
    mocker.patch(
        "management_commands.conf.settings.SUBMODULES",
        [
            "submodule_a",
            "submodule_b",
        ],
    )

    # Arrange.
    app_config_mock = mocker.Mock()
    app_config_mock.name = "app"

    # Mock.
    mocker.patch(
        "management_commands.core.apps.app_configs",
        {
            "app": app_config_mock,
        },
    )

    # Act.
    command_paths = get_command_paths("command", app_label="app")

    # Assert.
    assert command_paths == [
        "app.submodule_a.command.Command",
        "app.submodule_b.command.Command",
    ]


def test_get_command_path_raises_command_app_lookup_error_if_app_label_refer_to_not_installed_app(
    mocker: MockerFixture,
) -> None:
    # Mock.
    mocker.patch("management_commands.core.apps.app_configs", {})

    # Act & assert.
    with pytest.raises(CommandAppLookupError):
        get_command_paths("command", app_label="app")


def test_load_command_class_returns_command_class_if_any_of_dotted_paths_points_to_base_command_subclass(
    mocker: MockerFixture,
) -> None:
    # Configure.
    mocker.patch(
        "management_commands.conf.settings.SUBMODULES",
        [
            "submodule_a",
            "submodule_b",
            "submodule_c",
        ],
    )

    # Arrange.
    class CommandA:
        pass

    class CommandB(BaseCommand):
        pass

    class CommandC:
        pass

    app_config_mock = mocker.Mock()
    app_config_mock.name = "app"

    # Mock.
    def import_string_side_effect(dotted_path: str) -> type:
        if dotted_path == "app.submodule_a.command.Command":
            return CommandA
        if dotted_path == "app.submodule_b.command.Command":
            return CommandB
        if dotted_path == "app.submodule_c.command.Command":
            return CommandC
        raise ImportError

    mocker.patch(
        "management_commands.core.apps.app_configs",
        {
            "app": app_config_mock,
        },
    )
    mocker.patch(
        "management_commands.core.import_string",
        side_effect=import_string_side_effect,
    )

    # Act.
    command_class = load_command_class("command")

    # Assert.
    assert command_class is CommandB


def test_load_command_class_raises_command_class_lookup_error_if_none_of_dotted_paths_points_to_base_command_subclass(
    mocker: MockerFixture,
) -> None:
    # Mock.
    mocker.patch(
        "management_commands.core.import_string",
        side_effect=ImportError,
    )

    # Act & assert.
    with pytest.raises(CommandClassLookupError):
        load_command_class("command")



from __future__ import annotations

from contextlib import suppress

from django.apps.registry import apps
from django.core.management.base import BaseCommand
from django.utils.module_loading import import_string

from .conf import settings
from .exceptions import (
    CommandAppLookupError,
    CommandClassLookupError,
    CommandImportError,
    CommandTypeError,
)


def import_command_class(dotted_path: str) -> type[BaseCommand]:
    try:
        command_class: type = import_string(dotted_path)
    except ImportError as exc:
        raise CommandImportError(dotted_path) from exc

    if not (isinstance(command_class, type) and issubclass(command_class, BaseCommand)):
        raise CommandTypeError(command_class)

    return command_class


def get_command_paths(name: str, app_label: str | None = None) -> list[str]:
    if not app_label:
        app_names = [
            "django.core",
            *(app_config.name for app_config in reversed(list(apps.get_app_configs()))),
        ]

        modules_paths = [f"{module}.{name}.Command" for module in settings.MODULES]
    else:
        try:
            app_config = apps.get_app_config(app_label)
        except LookupError as exc:
            raise CommandAppLookupError(app_label) from exc
        else:
            app_names = [app_config.name]

        modules_paths = []

    submodules_paths: list[str] = []
    for app_name in app_names:
        for submodule in settings.SUBMODULES:
            if app_name == "django.core" and submodule != "management.commands":
                continue

            submodules_paths.append(f"{app_name}.{submodule}.{name}.Command")

    return modules_paths + submodules_paths


def load_command_class(name: str, app_label: str | None = None) -> type[BaseCommand]:
    command_paths = get_command_paths(name, app_label)

    for command_path in command_paths:
        with suppress(CommandImportError, CommandTypeError):
            return import_command_class(command_path)

    raise CommandClassLookupError(name, app_label)



from __future__ import annotations

import pytest

from management_commands.conf import settings


def test_configure_paths_raises_improperly_configured_with_invalid_command_key() -> None:  # fmt: skip
    # Arrange.
    paths = {
        "*command_a": "module_a.Command",
        "command_b": "module_b.Command",
    }

    # Act & assert.
    with pytest.raises(settings.ImproperlyConfigured) as exc_info:
        settings.configure_paths(paths)

    assert exc_info.value.code == "paths.key"


def test_configure_paths_raises_improperly_configured_with_invalid_command_value() -> None:  # fmt: skip
    # Arrange.
    paths = {
        "command_a": "module_a.Command",
        "command_b": "*module_b.Command",
    }

    # Act & assert.
    with pytest.raises(settings.ImproperlyConfigured) as exc_info:
        settings.configure_paths(paths)

    assert exc_info.value.code == "paths.value"


def test_configure_modules_raises_improperly_configured_with_invalid_module() -> None:
    # Arrange.
    modules = [
        "*module_a",
        "module_b",
    ]

    # Act & assert.
    with pytest.raises(settings.ImproperlyConfigured) as exc_info:
        settings.configure_modules(modules)

    assert exc_info.value.code == "modules.item"


def test_configure_submodules_prepends_management_commands_to_submodules_if_not_included() -> None:  # fmt: skip
    # Arrange.
    submodules = [
        "submodule_a",
        "submodule_b",
    ]

    # Act.
    configured_submodules = settings.configure_submodules(submodules)

    # Assert.
    assert configured_submodules == [
        "management.commands",
        "submodule_a",
        "submodule_b",
    ]


def test_configure_submodule_does_not_move_management_commands_if_already_included() -> None:  # fmt: skip
    # Arrange.
    submodules = [
        "submodule_a",
        "management.commands",
        "submodule_b",
    ]

    # Act.
    configured_submodules = settings.configure_submodules(submodules)

    # Assert.
    assert configured_submodules == submodules


def test_configure_submodules_raises_improperly_configured_with_invalid_module() -> None:  # fmt: skip
    # Arrange.
    submodules = [
        "*submodule_a",
        "submodule_b",
    ]

    # Act & assert.
    with pytest.raises(settings.ImproperlyConfigured) as exc_info:
        settings.configure_submodules(submodules)

    assert exc_info.value.code == "submodules.item"


def test_configure_aliases_raises_improperly_configured_with_invalid_key() -> None:
    # Arrange.
    aliases = {
        "*alias_a": ["command_a"],
        "alias_b": ["command_b"],
    }

    # Act & assert.
    with pytest.raises(settings.ImproperlyConfigured) as exc_info:
        settings.configure_aliases(aliases)

    assert exc_info.value.code == "aliases.key"


def test_configure_aliases_raises_improperly_configured_with_empty_command_expressions() -> None:  # fmt: skip
    # Arrange.
    aliases = {
        "alias": [""],
    }

    # Act & assert.
    with pytest.raises(settings.ImproperlyConfigured) as exc_info:
        settings.configure_aliases(aliases)

    assert exc_info.value.code == "aliases.empty"


def test_configure_aliases_raises_improperly_configured_with_self_reference() -> None:
    # Arrange.
    aliases = {
        "alias_a": ["alias_b"],
        "alias_b": ["alias_b"],
    }

    # Act & assert.
    with pytest.raises(settings.ImproperlyConfigured) as exc_info:
        settings.configure_aliases(aliases)

    assert exc_info.value.code == "aliases.self_reference"
